package com.muyu.domain.mysql;

import com.muyu.domain.mysql.config.MysqlPoolConfig;
import lombok.extern.log4j.Log4j2;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.util.Queue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.atomic.AtomicInteger;

/**
 * mysql连接池信息
 */
@Log4j2
public class MysqlPool implements BasePool<Connection> {


    /**
     * 等待(空闲队列) 基础队列
     */
    private Queue<Connection>  mysqlConneQueue = null;
    /**
     * 活动队列
     */
    private Queue<Connection> activeMysqlConnQueue = null;

    /**
     * 记录队列的连接总数
     */
    private AtomicInteger count=new AtomicInteger();

    //mysql基础配置
    public MysqlPoolConfig mysqlPoolConfig;
    //进行实例化连接池 加载驱动
    public MysqlPool(MysqlPoolConfig mysqlPoolConfig) {
        this.mysqlPoolConfig = mysqlPoolConfig;
        BaseConfig.driver(mysqlPoolConfig.getDriverName());
        log.info("MySQL连接池实例化完成");
    }


    //进行初始化连接池
    @Override
    public void init() {
        Long maxLinkNum = this.mysqlPoolConfig.getMaxLinkNum();
        Long initLinkNum = this.mysqlPoolConfig.getInitLinkNum();
        this.mysqlConneQueue=new LinkedBlockingQueue<Connection>(Math.toIntExact(maxLinkNum));
        this.activeMysqlConnQueue=new LinkedBlockingQueue<Connection>(Math.toIntExact(initLinkNum));
        for (int i = 0; i < initLinkNum; i++) {
            this.mysqlConneQueue.offer(createConn());
            count.incrementAndGet();
        }
        log.info("MYSQL连接池初始化完成");
    }


    /**
     * 1. 获取当前的毫秒，用来计算获取连接是否超时
     * 2. 判断队列当中是否还有空闲连接
     * 3. 如果有链接我们获取连接进行返回
     * 4. 如果没有 判断是否达到最大的连接数
     * 5. 如果达到最大连接数量进行等待（自旋）
     * 6. 如果没有达到最大连接数量进行创建连接
     * 7. 放到使用队列当中，染回在进行返回
     * 8.如果连接超时的话则进行抛出异常
     * @return
     */
    @Override
    public Connection getConn() {
        long startTime = System.currentTimeMillis();

        //从空闲队列当中取出放入活动队列
        Connection conn = this.mysqlConneQueue.poll();
        if(conn!=null){
            this.activeMysqlConnQueue.offer(conn);
            return conn;
        }
        //如果当前的连接数量小于最大的连接数量的时候创建新的连接0000.
        if(count.get()>this.mysqlPoolConfig.getMaxLinkNum()){
            Connection mysqlConn = createConn();
            this.activeMysqlConnQueue.offer(mysqlConn);
            count.incrementAndGet();
            return mysqlConn;
        }

        if((System.currentTimeMillis()-startTime)>this.mysqlPoolConfig.getMaxWaitTime()){
                throw new MysqlConnException("连接超时");
        }

        return null;
    }
    //归还连接
    @Override
    public void replease(Connection conn) {
        //删除活动队列当中的 连接
        if(this.mysqlConneQueue.remove(conn)){
            //把连接放到空闲队列当中
            this.mysqlConneQueue.offer(conn);
        }

    }
    //获取mysql连接信息
    public Connection createConn() {
        String url = this.mysqlPoolConfig.getUrl();
        String userName = this.mysqlPoolConfig.getUsername();
        String password = this.mysqlPoolConfig.getPassword();
        Connection mysqlConn = null;
        try {
            mysqlConn = DriverManager.getConnection(url, userName, password);
        } catch (SQLException e) {
            throw new RuntimeException(e);
        }
        log.info("初始化一个数据库连接");
        return mysqlConn;
    }

    //关闭连接
    @Override
    public void closeConn() {
        this.closeBaseConn();
        this.closeActiveConn();
    }
    //关闭空闲连接
    /**
     * 关闭空闲连接
     */
    public void closeBaseConn() {
        //从空闲连接当中拿出一个连接 准备进行关闭
        //如何拿出的这个链接为null 表示以列当中没有连接信息
        Connection poll = this.mysqlConneQueue.poll();
        if (poll!=null){
            try {
                poll.close();
            } catch (SQLException e) {
                try {
                    //判断这个接是否被关闭了，如果连接被关闭则不需要放入队列当中
                    //如何这个链接没有被关闭 则放入队列当中 尝试下次关闭
                    if (!poll.isClosed()){
                        this.mysqlConneQueue.offer(poll);
                    }
                } catch (SQLException ex) {
                    throw new RuntimeException(ex);
                }
                throw new RuntimeException(e);
            }finally {
                closeBaseConn();
            }

        }

    }

    //关闭活动队列
    public void closeActiveConn() {
        //从空闲连接当中拿出一个连接 准备进行关闭
        //如何拿出的这个链接为null 表示以列当中没有连接信息
        Connection poll = this.activeMysqlConnQueue.poll();
        if (poll!=null){
            try {
                poll.close();
            } catch (SQLException e) {
                try {
                    //判断这个接是否被关闭了，如果连接被关闭则不需要放入队列当中
                    //如何这个链接没有被关闭 则放入队列当中 尝试下次关闭
                    if (!poll.isClosed()){
                        this.activeMysqlConnQueue.offer(poll);
                    }
                } catch (SQLException ex) {
                    throw new RuntimeException(ex);
                }
                throw new RuntimeException(e);
            }finally {
                closeActiveConn();
            }
        }

    }




}
